import numpy as np
import pytransform3d.visualizer as pv
from movement_primitives.kinematics import Kinematics


with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    kin = Kinematics(f.read(), mesh_path="kuka_lbr/urdf/")
right_chain = kin.create_chain(
    ["kuka_lbr_r_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_r_tcp", verbose=0)
left_chain = kin.create_chain(
    ["kuka_lbr_l_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_l_tcp", verbose=0)
left_ee2base = left_chain.forward(np.array([0.5, 0.1, 0.0, -0.3, 0.0, 0.3, 0]) * np.pi)
right_ee2base = right_chain.forward(np.array([-0.5, 0.1, 0.0, -0.3, 0.0, 0.3, 0]) * np.pi)
right_chain.forward(np.zeros(7))
left_chain.forward(np.zeros(7))
random_state = np.random.RandomState(5)


def animation_callback(step, right_chain, left_chain, graph, right_ee2base, left_ee2base, random_state):
    right_chain.inverse_with_random_restarts(right_ee2base, random_state=random_state)
    left_chain.inverse_with_random_restarts(left_ee2base, random_state=random_state)
    graph.set_data()
    return graph


fig = pv.figure()
fig.plot_basis(s=0.1)
graph = fig.plot_graph(kin.tm, "kuka_lbr", show_visuals=True)
#graph = fig.plot_graph(kin.tm, "kuka_lbr", show_frames=True, s=0.1)
fig.view_init()
fig.animate(animation_callback, 1, loop=True, fargs=(right_chain, left_chain, graph, right_ee2base, left_ee2base, random_state))
fig.show()



import numpy as np


def generate_minimum_jerk(start, goal, execution_time=1.0, dt=0.01):
    """Create a minimum jerk trajectory.

    A minimum jerk trajectory from :math:`x_0` to :math:`g` minimizes
    the third time derivative of the positions:

    .. math::

        \\arg \min_{x_0, \ldots, x_T} \int_{t=0}^T \dddot{x}(t)^2 dt

    The trajectory will have

    .. code-block:: python

        n_steps = 1 + execution_time / dt

    steps because we start at 0 seconds and end at execution_time seconds.

    Parameters
    ----------
    start : array-like, shape (n_dims,)
        Initial state

    goal : array-like, shape (n_dims,)
        Goal state

    execution_time : float, optional (default: 1)
        Execution time in seconds

    dt : float, optional (default: 0.01)
        Time between successive steps in seconds

    Returns
    -------
    X : array, shape (n_steps, n_dims)
        The positions of the trajectory

    Xd : array, shape (n_steps, n_dims)
        The velocities of the trajectory

    Xdd : array, shape (n_steps, n_dims)
        The accelerations of the trajectory

    Raises
    ------
    ValueError
        If the shapes of the initial and goal state do not match.
    """
    x0 = np.asarray(start)
    g = np.asarray(goal)
    if x0.shape != g.shape:
        raise ValueError("Shape of initial state %s and goal %s must be equal"
                         % (x0.shape, g.shape))

    n_dims = x0.shape[0]
    n_steps = 1 + int(execution_time / dt)

    X = np.zeros((n_steps, n_dims))
    Xd = np.zeros((n_steps, n_dims))
    Xdd = np.zeros((n_steps, n_dims))

    x = x0.copy()
    xd = np.zeros(n_dims)
    xdd = np.zeros(n_dims)

    X[0] = x
    for t in range(1, n_steps):
        tau = execution_time - t * dt

        if tau >= dt:
            dist = g - x

            a1 = 0
            a0 = xdd * tau ** 2
            v1 = 0
            v0 = xd * tau

            t1 = dt
            t2 = dt ** 2
            t3 = dt ** 3
            t4 = dt ** 4
            t5 = dt ** 5

            c1 = (6. * dist + (a1 - a0) / 2. - 3. * (v0 + v1)) / tau ** 5
            c2 = (-15. * dist + (3. * a0 - 2. * a1) / 2. + 8. * v0 +
                  7. * v1) / tau ** 4
            c3 = (10. * dist + (a1 - 3. * a0) / 2. - 6. * v0 -
                  4. * v1) / tau ** 3
            c4 = xdd / 2.
            c5 = xd
            c6 = x

            x = c1 * t5 + c2 * t4 + c3 * t3 + c4 * t2 + c5 * t1 + c6
            xd = (5. * c1 * t4 + 4 * c2 * t3 + 3 * c3 * t2 + 2 * c4 * t1 + c5)
            xdd = (20. * c1 * t3 + 12. * c2 * t2 + 6. * c3 * t1 + 2. * c4)

        X[t] = x
        Xd[t] = xd
        Xdd[t] = xdd

    return X, Xd, Xdd



import math
import numpy as np
from scipy.interpolate import interp1d
import pytransform3d.rotations as pr
import pytransform3d.batch_rotations as pbr
import pytransform3d.transformations as pt


EPSILON = 1e-10

try:
    from ..dmp_fast import obstacle_avoidance_acceleration_2d as obstacle_avoidance_acceleration_2d_fast
    obstacle_avoidance_acceleration_2d_fast_available = True
except ImportError:
    obstacle_avoidance_acceleration_2d_fast_available = False


def obstacle_avoidance_acceleration_2d(
        y, yd, obstacle_position, gamma=1000.0, beta=20.0 / math.pi):
    """Compute acceleration for obstacle avoidance in 2D.

    Parameters
    ----------
    y : array, shape (..., 2)
        Current position(s).

    yd : array, shape (..., 2)
        Current velocity / velocities.

    obstacle_position : array, shape (2,)
        Position of the point obstacle.

    gamma : float, optional (default: 1000)
        Obstacle avoidance parameter.

    beta : float, optional (default: 20 / pi)
        Obstacle avoidance parameter.

    Returns
    -------
    cdd : array, shape (..., 2)
        Accelerations.
    """
    obstacle_diff = obstacle_position - y
    pad_width = ([[0, 0]] * (y.ndim - 1)) + [[0, 1]]
    obstacle_diff_0 = np.pad(obstacle_diff, pad_width, mode="constant",
                             constant_values=0.0)
    yd_0 = np.pad(yd, pad_width, mode="constant", constant_values=0.0)
    r = 0.5 * np.pi * pbr.norm_vectors(np.cross(obstacle_diff_0, yd_0))
    R = pbr.matrices_from_compact_axis_angles(r)[..., :2, :2]
    theta_nom = np.einsum(
        "ni,ni->n", obstacle_diff.reshape(-1, 2), yd.reshape(-1, 2))
    shape = y.shape[:-1]
    if shape:
        theta_nom = theta_nom.reshape(*shape)
    theta_denom = (np.linalg.norm(obstacle_diff, axis=-1)
                   * np.linalg.norm(yd, axis=-1) + EPSILON)
    theta = np.arccos(theta_nom / theta_denom)
    rotated_velocity = np.einsum(
        "nij,nj->ni", R.reshape(-1, 2, 2), yd.reshape(-1, 2))
    if shape:
        rotated_velocity = rotated_velocity.reshape(*(shape + (2,)))
    cdd = (gamma * rotated_velocity
           * (theta * np.exp(-beta * theta))[..., np.newaxis])
    return np.squeeze(cdd)


class CouplingTermObstacleAvoidance2D:
    """Coupling term for obstacle avoidance in 2D.

    For :class:`DMP` and :class:`DMPWithFinalVelocity`.

    This is the simplified 2D implementation of
    :class:`CouplingTermObstacleAvoidance3D`.

    Parameters
    ----------
    obstacle_position : array, shape (2,)
        Position of the point obstacle.

    gamma : float, optional (default: 1000)
        Parameter of obstacle avoidance.

    beta : float, optional (default: 20 / pi)
        Parameter of obstacle avoidance.
    """
    def __init__(self, obstacle_position, gamma=1000.0, beta=20.0 / math.pi,
                 fast=False):
        self.obstacle_position = obstacle_position
        self.gamma = gamma
        self.beta = beta
        if fast and obstacle_avoidance_acceleration_2d_fast_available:
            self.step_function = obstacle_avoidance_acceleration_2d_fast
        else:
            self.step_function = obstacle_avoidance_acceleration_2d

    def coupling(self, y, yd):
        """Computes coupling term based on current state.

        Parameters
        ----------
        y : array, shape (n_dims,)
            Current position.

        yd : array, shape (n_dims,)
            Current velocity.

        Returns
        -------
        cd : array, shape (n_dims,)
            Velocity. 0 for this coupling term.

        cdd : array, shape (n_dims,)
            Acceleration.
        """
        cdd = self.step_function(
            y, yd, self.obstacle_position, self.gamma, self.beta)
        return np.zeros_like(cdd), cdd


class CouplingTermObstacleAvoidance3D:  # for DMP
    r"""Coupling term for obstacle avoidance in 3D.

    For :class:`DMP` and :class:`DMPWithFinalVelocity`.

    Following [1]_, this coupling term adds an acceleration

    .. math::

        \boldsymbol{C}_t = \gamma \boldsymbol{R} \dot{\boldsymbol{y}}
        \theta \exp(-\beta \theta),

    where

    .. math::

        \theta = \arccos\left( \frac{(\boldsymbol{o} - \boldsymbol{y})^T
        \dot{\boldsymbol{y}}}{|\boldsymbol{o} - \boldsymbol{y}|
        |\dot{\boldsymbol{y}}|} \right)

    and a rotation axis :math:`\boldsymbol{r} =
    (\boldsymbol{o} - \boldsymbol{y}) \times \dot{\boldsymbol{y}}` used to
    compute the rotation matrix :math:`\boldsymbol{R}` that rotates about it
    by 90 degrees for an obstacle at position :math:`\boldsymbol{o}`.

    Intuitively, this coupling term adds a movement perpendicular to the
    current velocity in the plane defined by the direction to the obstacle and
    the current movement direction.

    Parameters
    ----------
    obstacle_position : array, shape (3,)
        Position of the point obstacle.

    gamma : float, optional (default: 1000)
        Parameter of obstacle avoidance.

    beta : float, optional (default: 20 / pi)
        Parameter of obstacle avoidance.

    References
    ----------
    .. [1] Ijspeert, A. J., Nakanishi, J., Hoffmann, H., Pastor, P., Schaal, S.
       (2013). Dynamical Movement Primitives: Learning Attractor Models for
       Motor Behaviors. Neural Computation 25 (2), 328-373. DOI:
       10.1162/NECO_a_00393,
       https://homes.cs.washington.edu/~todorov/courses/amath579/reading/DynamicPrimitives.pdf
    """
    def __init__(self, obstacle_position, gamma=1000.0, beta=20.0 / math.pi):
        self.obstacle_position = obstacle_position
        self.gamma = gamma
        self.beta = beta

    def coupling(self, y, yd):
        """Computes coupling term based on current state.

        Parameters
        ----------
        y : array, shape (n_dims,)
            Current position.

        yd : array, shape (n_dims,)
            Current velocity.

        Returns
        -------
        cd : array, shape (n_dims,)
            Velocity. 0 for this coupling term.

        cdd : array, shape (n_dims,)
            Acceleration.
        """
        cdd = obstacle_avoidance_acceleration_3d(
            y, yd, self.obstacle_position, self.gamma, self.beta)
        return np.zeros_like(cdd), cdd


def obstacle_avoidance_acceleration_3d(
        y, yd, obstacle_position, gamma=1000.0, beta=20.0 / math.pi):
    """Compute acceleration for obstacle avoidance in 3D.

    Parameters
    ----------
    y : array, shape (..., 3)
        Current position(s).

    yd : array, shape (..., 3)
        Current velocity / velocities.

    obstacle_position : array, shape (3,)
        Position of the point obstacle.

    gamma : float, optional (default: 1000)
        Obstacle avoidance parameter.

    beta : float, optional (default: 20 / pi)
        Obstacle avoidance parameter.

    Returns
    -------
    cdd : array, shape (..., 3)
        Accelerations.
    """
    obstacle_diff = obstacle_position - y
    r = 0.5 * np.pi * pr.norm_vector(np.cross(obstacle_diff, yd))
    R = pr.matrix_from_compact_axis_angle(r)
    theta = np.arccos(
        np.dot(obstacle_diff, yd)
        / (np.linalg.norm(obstacle_diff) * np.linalg.norm(yd) + EPSILON))
    cdd = gamma * np.dot(R, yd) * theta * np.exp(-beta * theta)
    return cdd


class CouplingTermPos1DToPos1D:
    """Couples position components of a 2D DMP with a virtual spring.

    For :class:`DMP` and :class:`DMPWithFinalVelocity`.

    Parameters
    ----------
    desired_distance : float
        Desired distance between components.

    lf : array-like, shape (2,)
        Binary values that indicate which DMP(s) will be adapted.
        The variable lf defines the relation leader-follower. If lf[0] = lf[1],
        then both robots will adapt their trajectories to follow average
        trajectories at the defined distance dd between them [..]. On the other
        hand, if lf[0] = 0 and lf[1] = 1, only DMP1 will change the trajectory
        to match the trajectory of DMP0, again at the distance dd and again
        only after learning. Vice versa applies as well. Leader-follower
        relation can be determined by a higher-level planner [..].

    k : float, optional (default: 1)
        Virtual spring constant that couples the positions.

    c1 : float, optional (default: 100)
        Scaling factor for spring forces in the velocity component and
        acceleration component.

    c2 : float, optional (default: 30)
        Scaling factor for spring forces in the acceleration component.

    References
    ----------
    .. [1] Gams, A., Nemec, B., Zlajpah, L., Wächter, M., Asfour, T., Ude, A.
       (2013). Modulation of Motor Primitives using Force Feedback: Interaction
       with the Environment and Bimanual Tasks (2013), In 2013 IEEE/RSJ
       International Conference on Intelligent Robots and Systems (pp.
       5629-5635). DOI: 10.1109/IROS.2013.6697172,
       https://h2t.anthropomatik.kit.edu/pdf/Gams2013.pdf
    """
    def __init__(self, desired_distance, lf, k=1.0, c1=100.0, c2=30.0):
        self.desired_distance = desired_distance
        self.lf = lf
        self.k = k
        self.c1 = c1
        self.c2 = c2

    def coupling(self, y, yd=None):
        da = y[0] - y[1]
        F12 = self.k * (-self.desired_distance - da)
        F21 = -F12
        C12 = self.c1 * F12 * self.lf[0]
        C21 = self.c1 * F21 * self.lf[1]
        C12dot = self.c2 * self.c1 * F12 * self.lf[0]
        C21dot = self.c2 * self.c1 * F21 * self.lf[1]
        return np.array([C12, C21]), np.array([C12dot, C21dot])


class CouplingTermPos3DToPos3D:
    """Couples position components of a 6D DMP with a virtual spring in 3D.

    For :class:`DMP` and :class:`DMPWithFinalVelocity`.

    Parameters
    ----------
    desired_distance : array, shape (3,)
        Desired distance between components.

    lf : array-like, shape (2,)
        Binary values that indicate which DMP(s) will be adapted.
        The variable lf defines the relation leader-follower. If lf[0] = lf[1],
        then both robots will adapt their trajectories to follow average
        trajectories at the defined distance dd between them [..]. On the other
        hand, if lf[0] = 0 and lf[1] = 1, only DMP1 will change the trajectory
        to match the trajectory of DMP0, again at the distance dd and again
        only after learning. Vice versa applies as well. Leader-follower
        relation can be determined by a higher-level planner [..].

    k : float, optional (default: 1)
        Virtual spring constant that couples the positions.

    c1 : float, optional (default: 100)
        Scaling factor for spring forces in the velocity component and
        acceleration component.

    c2 : float, optional (default: 30)
        Scaling factor for spring forces in the acceleration component.

    References
    ----------
    .. [1] Gams, A., Nemec, B., Zlajpah, L., Wächter, M., Asfour, T., Ude, A.
       (2013). Modulation of Motor Primitives using Force Feedback: Interaction
       with the Environment and Bimanual Tasks (2013), In 2013 IEEE/RSJ
       International Conference on Intelligent Robots and Systems (pp.
       5629-5635). DOI: 10.1109/IROS.2013.6697172,
       https://h2t.anthropomatik.kit.edu/pdf/Gams2013.pdf
    """
    def __init__(self, desired_distance, lf, k=1.0, c1=1.0, c2=30.0):
        self.desired_distance = desired_distance
        self.lf = lf
        self.k = k
        self.c1 = c1
        self.c2 = c2

    def coupling(self, y, yd=None):
        da = y[:3] - y[3:6]
        # Why do we take -self.desired_distance here? Because this allows us
        # to regard the desired distance as the displacement of DMP1 with
        # respect to DMP0.
        F12 = self.k * (-self.desired_distance - da)
        F21 = -F12
        C12 = self.c1 * F12 * self.lf[0]
        C21 = self.c1 * F21 * self.lf[1]
        C12dot = F12 * self.c2 * self.lf[0]
        C21dot = F21 * self.c2 * self.lf[1]
        return np.hstack([C12, C21]), np.hstack([C12dot, C21dot])


class CouplingTermDualCartesianDistance:
    """Couples distance between 3D positions of a dual Cartesian DMP.

    For :class:`DualCartesianDMP`.

    Parameters
    ----------
    desired_distance : float
        Desired distance between components.

    lf : array-like, shape (2,)
        Binary values that indicate which DMP(s) will be adapted.
        The variable lf defines the relation leader-follower. If lf[0] = lf[1],
        then both robots will adapt their trajectories to follow average
        trajectories at the defined distance dd between them [..]. On the other
        hand, if lf[0] = 0 and lf[1] = 1, only DMP1 will change the trajectory
        to match the trajectory of DMP0, again at the distance dd and again
        only after learning. Vice versa applies as well. Leader-follower
        relation can be determined by a higher-level planner [..].

    k : float, optional (default: 1)
        Virtual spring constant that couples the positions.

    c1 : float, optional (default: 100)
        Scaling factor for spring forces in the velocity component and
        acceleration component.

    c2 : float, optional (default: 30)
        Scaling factor for spring forces in the acceleration component.
    """
    def __init__(self, desired_distance, lf, k=1.0, c1=1.0, c2=30.0):
        self.desired_distance = desired_distance
        self.lf = lf
        self.k = k
        self.c1 = c1
        self.c2 = c2

    def coupling(self, y, yd=None):
        actual_distance = y[:3] - y[7:10]
        desired_distance = (np.abs(self.desired_distance) * actual_distance
                            / np.linalg.norm(actual_distance))
        F12 = self.k * (desired_distance - actual_distance)
        F21 = -F12
        C12 = self.c1 * F12 * self.lf[0]
        C21 = self.c1 * F21 * self.lf[1]
        C12dot = F12 * self.c2 * self.lf[0]
        C21dot = F21 * self.c2 * self.lf[1]
        return (np.hstack([C12, np.zeros(3), C21, np.zeros(3)]),
                np.hstack([C12dot, np.zeros(3), C21dot, np.zeros(3)]))


class CouplingTermDualCartesianPose:
    """Couples relative poses of dual Cartesian DMP.

    For :class:`DualCartesianDMP`.

    Parameters
    ----------
    desired_distance : array, shape (4, 4)
        Desired distance between components.

    lf : array-like, shape (2,)
        Binary values that indicate which DMP(s) will be adapted.
        The variable lf defines the relation leader-follower. If lf[0] = lf[1],
        then both robots will adapt their trajectories to follow average
        trajectories at the defined distance dd between them [..]. On the other
        hand, if lf[0] = 0 and lf[1] = 1, only DMP1 will change the trajectory
        to match the trajectory of DMP0, again at the distance dd and again
        only after learning. Vice versa applies as well. Leader-follower
        relation can be determined by a higher-level planner [..].

    couple_position : bool, optional (default: True)
        Couple position between components.

    couple_orientation : bool, optional (default: True)
        Couple orientation between components.

    k : float, optional (default: 1)
        Virtual spring constant that couples the positions.

    c1 : float, optional (default: 100)
        Scaling factor for spring forces in the velocity component and
        acceleration component.

    c2 : float, optional (default: 30)
        Scaling factor for spring forces in the acceleration component.
    """
    def __init__(self, desired_distance, lf, couple_position=True,
                 couple_orientation=True, k=1.0, c1=1.0, c2=30.0, verbose=0):
        self.desired_distance = desired_distance
        self.lf = lf
        self.couple_position = couple_position
        self.couple_orientation = couple_orientation
        self.k = k
        self.c1 = c1
        self.c2 = c2
        self.verbose = verbose

    def coupling(self, y, yd=None):
        return self.couple_distance(
            y, yd, self.k, self.c1, self.c2, self.lf, self.desired_distance,
            self.couple_position, self.couple_orientation)

    def couple_distance(self, y, yd, k, c1, c2, lf, desired_distance,
                        couple_position, couple_orientation):
        damping = 2.0 * np.sqrt(k * c2)

        vel_left = yd[:6]
        vel_right = yd[6:]

        left2base = pt.transform_from_pq(y[:7])
        right2left_pq = self._right2left_pq(y)

        actual_distance_pos = right2left_pq[:3]
        actual_distance_rot = right2left_pq[3:]

        desired_distance = pt.pq_from_transform(desired_distance)
        desired_distance_pos = desired_distance[:3]
        desired_distance_rot = desired_distance[3:]

        if self.verbose:
            print("Desired vs. actual:")
            print(np.round(desired_distance, 2))
            print(np.round(right2left_pq, 2))

        error_pos = desired_distance_pos - actual_distance_pos
        F12_pos = -k * error_pos
        F21_pos = k * error_pos

        F12_pos = pt.transform(left2base, pt.vector_to_direction(F12_pos))[:3]
        F21_pos = pt.transform(left2base, pt.vector_to_direction(F21_pos))[:3]

        C12_pos = lf[0] * c1 * F12_pos
        C21_pos = lf[1] * c1 * F21_pos

        C12dot_pos = lf[0] * (c2 * F12_pos - damping * vel_left[:3])
        C21dot_pos = lf[1] * (c2 * F21_pos - damping * vel_right[:3])

        if not couple_position:
            C12_pos *= 0
            C21_pos *= 0
            C12dot_pos *= 0
            C21dot_pos *= 0

        error_rot = pr.compact_axis_angle_from_quaternion(
            pr.concatenate_quaternions(desired_distance_rot,
                                       pr.q_conj(actual_distance_rot)))
        F12_rot = -k * error_rot
        F21_rot = k * error_rot

        F12_rot = pt.transform(left2base, pt.vector_to_direction(F12_rot))[:3]
        F21_rot = pt.transform(left2base, pt.vector_to_direction(F21_rot))[:3]

        C12_rot = lf[0] * c1 * F12_rot
        C21_rot = lf[1] * c1 * F21_rot

        C12dot_rot = lf[0] * (c2 * F12_rot - damping * vel_left[3:])
        C21dot_rot = lf[1] * (c2 * F21_rot - damping * vel_right[3:])

        if not couple_orientation:
            C12_rot *= 0
            C21_rot *= 0
            C12dot_rot *= 0
            C21dot_rot *= 0

        return (np.hstack([C12_pos, C12_rot, C21_pos, C21_rot]),
                np.hstack([C12dot_pos, C12dot_rot, C21dot_pos, C21dot_rot]))

    def _right2left_pq(self, y):
        left2base = pt.transform_from_pq(y[:7])
        right2base = pt.transform_from_pq(y[7:])
        base2left = pt.invert_transform(left2base)
        right2left = pt.concat(right2base, base2left)
        right2left_pq = pt.pq_from_transform(right2left)
        return right2left_pq


class CouplingTermDualCartesianTrajectory(CouplingTermDualCartesianPose):
    """Couples relative pose in dual Cartesian DMP with a given trajectory.

    For :class:`DualCartesianDMP`.

    Parameters
    ----------
    offset : array, shape (7,)
        Offset for desired distance between components as position and
        quaternion.

    lf : array-like, shape (2,)
        Binary values that indicate which DMP(s) will be adapted.
        The variable lf defines the relation leader-follower. If lf[0] = lf[1],
        then both robots will adapt their trajectories to follow average
        trajectories at the defined distance dd between them [..]. On the other
        hand, if lf[0] = 0 and lf[1] = 1, only DMP1 will change the trajectory
        to match the trajectory of DMP0, again at the distance dd and again
        only after learning. Vice versa applies as well. Leader-follower
        relation can be determined by a higher-level planner [..].

    couple_position : bool, optional (default: True)
        Couple position between components.

    couple_orientation : bool, optional (default: True)
        Couple orientation between components.

    k : float, optional (default: 1)
        Virtual spring constant that couples the positions.

    c1 : float, optional (default: 100)
        Scaling factor for spring forces in the velocity component and
        acceleration component.

    c2 : float, optional (default: 30)
        Scaling factor for spring forces in the acceleration component.
    """
    def __init__(self, offset, lf, dt, couple_position=True,
                 couple_orientation=True, k=1.0, c1=1.0, c2=30.0, verbose=1):
        self.offset = offset
        self.lf = lf
        self.dt = dt
        self.couple_position = couple_position
        self.couple_orientation = couple_orientation
        self.k = k
        self.c1 = c1
        self.c2 = c2
        self.verbose = verbose

    def imitate(self, T, Y):
        distance = np.empty((len(Y), 7))
        for t in range(len(Y)):
            distance[t] = self._right2left_pq(Y[t])
        self.desired_distance_per_dimension = [
            interp1d(T, distance[:, d], bounds_error=False,
                     fill_value="extrapolate")
            for d in range(distance.shape[1])
        ]
        self.t = 0.0

    def coupling(self, y, yd=None):
        desired_distance = np.empty(len(self.desired_distance_per_dimension))
        for d in range(len(desired_distance)):
            desired_distance[d] = self.desired_distance_per_dimension[d](self.t)
        desired_distance += self.offset
        self.t += self.dt
        return self.couple_distance(
            y, yd, self.k, self.c1, self.c2, self.lf,
            pt.transform_from_pq(desired_distance), self.couple_position,
            self.couple_orientation)



"""
========================
Obstacle Avoidance in 2D
========================

Plots a 2D DMP that goes through a point obstacle when there is no coupling
term for obstacle avoidance and a 2D DMP that avoids the point obstacle with
a coupling term.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMP, CouplingTermObstacleAvoidance2D


execution_time = 1.0
start_y = np.zeros(2)
goal_y = np.ones(2)

dmp = DMP(n_dims=2, execution_time=execution_time, n_weights_per_dim=3)
dmp.configure(start_y=start_y, goal_y=goal_y)
dmp.set_weights(np.array([-50.0, 100.0, 300.0, -200.0, -200.0, -200.0]))

fig = plt.figure()
ax = fig.add_subplot(111)
ax.set_xlabel("x")
ax.set_ylabel("y")
obstacle_position = np.array([0.92, 0.5])
T, Y = dmp.open_loop(run_t=execution_time)
ax.plot(Y[:, 0], Y[:, 1], label="Original")
coupling_term = CouplingTermObstacleAvoidance2D(obstacle_position)
T, Y = dmp.open_loop(run_t=execution_time, coupling_term=coupling_term)
ax.plot(Y[:, 0], Y[:, 1], label="Obstacle avoidance")
ax.scatter(start_y[0], start_y[1], c="r", label="Start")
ax.scatter(goal_y[0], goal_y[1], c="g", label="Goal")
ax.scatter(obstacle_position[0], obstacle_position[1], c="y", label="Obstacle")
ax.legend()
plt.tight_layout()
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import time


sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = "movement_primitives"
copyright = "2020-{}, Alexander Fabisch, DFKI GmbH, Robotics Innovation Center".format(time.strftime("%Y"))
author = "Alexander Fabisch"

# The full version, including alpha/beta/rc tags
release = __import__(project).__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "numpydoc",
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autosummary_generate = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "logo": {
        "text": f"movement_primitives {release}"
    },
    "navbar_start": ["navbar-logo"],
}

root_doc = "index"
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_show_sourcelink = False
html_show_sphinx = False

intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(sys.version_info), None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'matplotlib': ('https://matplotlib.org/', None)
}



"""
========================
Simulate a Cartesian DMP
========================

A Cartesian DMP is used to represent a Cartesian trajectory given by positions
and quaternions.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import CartesianDMP
from movement_primitives.testing.simulation import UR5Simulation
from pytransform3d import rotations as pr


dt = 0.001
execution_time = 1.0

dmp = CartesianDMP(
    execution_time=execution_time, dt=dt,
    n_weights_per_dim=10, int_dt=0.0001)
Y = np.zeros((1001, 7))
T = np.linspace(0, 1, len(Y))
sigmoid = 0.5 * (np.tanh(1.5 * np.pi * (T - 0.5)) + 1.0)
Y[:, 0] = 0.6
Y[:, 1] = -0.2 + 0.4 * sigmoid
Y[:, 2] = 0.45
start_aa = np.array([0.0, 1.0, 0.0, 0.25 * np.pi])
goal_aa = np.array([0.0, 0.0, 1.0, 0.25 * np.pi])
for t in range(len(Y)):
    frac = sigmoid[t]
    aa_t = (1.0 - frac) * start_aa + frac * goal_aa
    aa_t[:3] /= np.linalg.norm(aa_t[:3])
    Y[t, 3:] = pr.quaternion_from_axis_angle(aa_t)
dmp.imitate(T, Y, allow_final_velocity=True)
dmp.configure(start_y=Y[0], goal_y=Y[-1])

ur5 = UR5Simulation(dt=dt, real_time=False)
for _ in range(4):
    ur5.goto_ee_state(Y[0], wait_time=1.0)
    ur5.stop()

desired_positions, positions, desired_velocities, velocities = \
    ur5.step_through_cartesian(dmp, Y[0], np.zeros(6), execution_time)

P = np.asarray(positions)
dP = np.asarray(desired_positions)
V = np.asarray(velocities)
dV = np.asarray(desired_velocities)

plot_dim = 5
plt.plot(T, Y[:, plot_dim], label="Demo")
plt.scatter([[0, T[-1]]], [[Y[0, plot_dim], Y[-1, plot_dim]]])
plt.plot(np.linspace(0, execution_time, len(P)), P[:, plot_dim], label="Actual")
plt.scatter([[0, execution_time]], [[P[0, plot_dim], P[-1, plot_dim]]])
plt.plot(np.linspace(0, execution_time, len(dP)), dP[:, plot_dim], label="Desired")
plt.scatter([[0, execution_time]], [[dP[0, plot_dim], dP[-1, plot_dim]]])
T, Y = dmp.open_loop(run_t=2.0)
plt.plot(T, Y[:, plot_dim], label="Open loop")
plt.scatter([[0, T[-1]]], [[Y[0, plot_dim], Y[-1, plot_dim]]])
plt.legend()
plt.show()



import numpy as np
import matplotlib.pyplot as plt
from pytransform3d.plot_utils import make_3d_axis
import pytransform3d.trajectories as ptr
from mocap.dataset_loader import load_rh5_demo
from movement_primitives.dmp import DualCartesianDMP, CouplingTermDualCartesianTrajectory


path = "data/rh5/20200831-1541/csv_processed/dual_arm_anticlockwise/20200831-1541_2.csv"
T, P = load_rh5_demo(path)
execution_time = T[-1]
dt = np.mean(np.diff(T))

dmp = DualCartesianDMP(
    execution_time=execution_time, dt=dt,
    n_weights_per_dim=10, int_dt=0.001, p_gain=0.0)
dmp.imitate(T, P)
dmp.configure(start_y=P[0], goal_y=P[-1])
offset = np.array([-0.05, 0.05, 0, 1, 0, 0, 0])
ct = CouplingTermDualCartesianTrajectory(
    offset=offset, dt=dt, couple_position=True, couple_orientation=False,
    lf=(1.0, 0.0), k=1, c1=0.1, c2=100)  # c2=10000 in simulation
ct.imitate(T, P)
T2, P2 = dmp.open_loop(run_t=execution_time, coupling_term=ct)

ax = make_3d_axis(ax_s=1)
ax.set_xlim((0.4, 1))
ax.set_ylim((-0.3, 0.3))
ax.set_zlim((0, 0.6))
ptr.plot_trajectory(ax=ax, P=P[:, :7], s=0.02, alpha=0.2)
ptr.plot_trajectory(ax=ax, P=P[:, 7:], s=0.02, alpha=0.2)
ptr.plot_trajectory(ax=ax, P=P2[:, :7], s=0.01)
ptr.plot_trajectory(ax=ax, P=P2[:, 7:], s=0.01)
plt.show()



"""
==================
Effect of DMP Gain
==================

Demonstrates how modifying DMP gains (alpha_y, beta_y) affects 
the resulting trajectory reproduction.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMP


dt = 0.01

dmp1 = DMP(
    n_dims=2,
    execution_time=1.0,
    dt=dt,
    n_weights_per_dim=10,
    int_dt=0.0001,
    alpha_y=np.array([25.0, 25.0]),
    beta_y=np.array([6.25, 6.25]),
)

dmp2 = DMP(
    n_dims=2,
    execution_time=1.0,
    dt=dt,
    n_weights_per_dim=10,
    int_dt=0.0001,
    alpha_y=np.array([25.0, 10.0]),  # note different alpha_y
    beta_y=np.array([6.25, 3.0]),  # note different beta_y
)

T = np.linspace(0.0, 1.0, 101)
Y = np.empty((101, 2))
Y[:, 0] = np.cos(np.pi * T)
Y[:, 1] = np.sin(np.pi * T)

plt.plot(Y[:, 0], Y[:, 1], label="Demo")
dmps = [dmp1, dmp2]
for i, dmp in enumerate(dmps):

    dmp.imitate(T, Y)
    dmp.configure(start_y=Y[0], goal_y=Y[-1])
    _, Y_ = dmp.open_loop()
    plt.plot(Y_[:, 0], Y_[:, 1], label=f"Reproduction {i+1}")

plt.grid()
plt.gca().set_aspect("equal", "box")
plt.legend()
plt.xlabel("x")
plt.ylabel("y")

plt.show()



"""Probabilistic Movement Primitive (ProMP)
========================================

ProMPs represent distributions of trajectories.
"""
import numpy as np


class ProMP:
    """Probabilistic Movement Primitive (ProMP).

    ProMPs have been proposed first in [1]_ and have been used later in [2]_,
    [3]_. The learning algorithm is a specialized form of the one presented in
    [4]_.

    Note that internally we represented trajectories with the task space
    dimension as the first axis and the time step as the second axis while
    the exposed trajectory interface is transposed. In addition, we internally
    only use a 1d array representation to make handling of the covariance
    simpler.

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    References
    ----------
    .. [1] Paraschos, A., Daniel, C., Peters, J., Neumann, G. (2013).
       Probabilistic movement primitives, In C.J. Burges and L. Bottou and M.
       Welling and Z. Ghahramani and K.Q. Weinberger (Eds.), Advances in Neural
       Information Processing Systems, 26,
       https://papers.nips.cc/paper/2013/file/e53a0a2978c28872a4505bdb51db06dc-Paper.pdf

    .. [3] Maeda, G. J., Neumann, G., Ewerton, M., Lioutikov, R., Kroemer, O.,
       Peters, J. (2017). Probabilistic movement primitives for coordination of
       multiple human–robot collaborative tasks. Autonomous Robots, 41, 593-612.
       DOI: 10.1007/s10514-016-9556-2,
       https://link.springer.com/article/10.1007/s10514-016-9556-2

    .. [2] Paraschos, A., Daniel, C., Peters, J., Neumann, G. (2018).
       Using probabilistic movement primitives in robotics. Autonomous Robots,
       42, 529-551. DOI: 10.1007/s10514-017-9648-7,
       https://www.ias.informatik.tu-darmstadt.de/uploads/Team/AlexandrosParaschos/promps_auro.pdf

    .. [4] Lazaric, A., Ghavamzadeh, M. (2010).
       Bayesian Multi-Task Reinforcement Learning. In Proceedings of the 27th
       International Conference on International Conference on Machine Learning
       (ICML'10) (pp. 599-606). https://hal.inria.fr/inria-00475214/document
    """
    def __init__(self, n_dims, n_weights_per_dim=10):
        self.n_dims = n_dims
        self.n_weights_per_dim = n_weights_per_dim

        self.n_weights = n_dims * n_weights_per_dim

        self.weight_mean = np.zeros(self.n_weights)
        self.weight_cov = np.eye(self.n_weights)

        self.centers = np.linspace(0, 1, self.n_weights_per_dim)

    def weights(self, T, Y, lmbda=1e-12):
        """Obtain ProMP weights by linear regression.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Y : array-like, shape (n_steps, n_dims)
            Demonstrated trajectory

        lmbda : float, optional (default: 1e-12)
            Regularization coefficient

        Returns
        -------
        weights : array, shape (n_steps * n_weights_per_dim)
            ProMP weights
        """
        activations = self._rbfs_nd_sequence(T).T
        weights = np.linalg.pinv(
            activations.T.dot(activations)
            + lmbda * np.eye(activations.shape[1])
        ).dot(activations.T).dot(Y.T.ravel())
        return weights

    def trajectory_from_weights(self, T, weights):
        """Generate trajectory from ProMP weights.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        weights : array-like, shape (n_steps * n_weights_per_dim)
            ProMP weights

        Returns
        -------
        Y : array, shape (n_steps, n_dims)
            Trajectory
        """
        return self._rbfs_nd_sequence(T).T.dot(weights).reshape(
            self.n_dims, len(T)).T

    def condition_position(self, y_mean, y_cov=None, t=0, t_max=1.0):
        """Condition ProMP on a specific position.

        For details, see page 4 of [1]_

        Parameters
        ----------
        y_mean : array, shape (n_dims,)
            Position mean.

        y_cov : array, shape (n_dims, n_dims), optional (default: 0)
            Covariance of position.

        t : float, optional (default: 0)
            Time at which the activations of RBFs will be queried. Note that
            we internally normalize the time so that t_max == 1.

        t_max : float, optional (default: 1)
            Duration of the ProMP

        Returns
        -------
        conditional_promp : ProMP
            New conditional ProMP

        References
        ----------
        .. [1] Paraschos, A., Daniel, C., Peters, J., Neumann, G. (2013).
           Probabilistic movement primitives, In C.J. Burges and L. Bottou and
           M. Welling and Z. Ghahramani and K.Q. Weinberger (Eds.), Advances in
           Neural Information Processing Systems, 26,
           https://papers.nips.cc/paper/2013/file/e53a0a2978c28872a4505bdb51db06dc-Paper.pdf
        """
        Psi_t = _nd_block_diagonal(
            self._rbfs_1d_point(t, t_max)[:, np.newaxis], self.n_dims)
        if y_cov is None:
            y_cov = 0.0

        common_term = self.weight_cov.dot(Psi_t).dot(
            np.linalg.inv(y_cov + Psi_t.T.dot(self.weight_cov).dot(Psi_t)))

        # Equation (5)
        weight_mean = (
            self.weight_mean
            + common_term.dot(y_mean - Psi_t.T.dot(self.weight_mean)))
        # Equation (6)
        weight_cov = (
            self.weight_cov - common_term.dot(Psi_t.T).dot(self.weight_cov))

        conditional_promp = ProMP(self.n_dims, self.n_weights_per_dim)
        conditional_promp.from_weight_distribution(weight_mean, weight_cov)
        return conditional_promp

    def mean_trajectory(self, T):
        """Get mean trajectory of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        Y : array, shape (n_steps, n_dims)
            Mean trajectory
        """
        return self.trajectory_from_weights(T, self.weight_mean)

    def cov_trajectory(self, T):
        """Get trajectory covariance of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        cov : array, shape (n_dims * n_steps, n_dims * n_steps)
            Covariance
        """
        activations = self._rbfs_nd_sequence(T)
        return activations.T.dot(self.weight_cov).dot(activations)

    def var_trajectory(self, T):
        """Get trajectory variance of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        var : array, shape (n_steps, n_dims)
            Variance
        """
        return np.maximum(np.diag(self.cov_trajectory(T)).reshape(
            self.n_dims, len(T)).T, 0.0)

    def mean_velocities(self, T):
        """Get mean velocities of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        Yd : array, shape (n_steps, n_dims)
            Mean velocities
        """
        return self._rbfs_derivative_nd_sequence(
            T).T.dot(self.weight_mean).reshape(self.n_dims, len(T)).T

    def cov_velocities(self, T):
        """Get velocity covariance of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        cov : array, shape (n_dims * n_steps, n_dims * n_steps)
            Covariance
        """
        activations = self._rbfs_derivative_nd_sequence(T)
        return activations.T.dot(self.weight_cov).dot(activations)

    def var_velocities(self, T):
        """Get velocity variance of ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        Returns
        -------
        var : array, shape (n_steps, n_dims)
            Variance
        """
        return np.maximum(np.diag(self.cov_velocities(T)).reshape(
            self.n_dims, len(T)).T, 0.0)

    def sample_trajectories(self, T, n_samples, random_state):
        """Sample trajectories from ProMP.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Time steps

        n_samples : int
            Number of trajectories that will be sampled

        random_state : np.random.RandomState
            State of random number generator

        Returns
        -------
        samples : array, shape (n_samples, n_steps, n_dims)
            Sampled trajectories
        """
        weight_samples = random_state.multivariate_normal(
            self.weight_mean, self.weight_cov, n_samples)
        samples = np.empty((n_samples, len(T), self.n_dims))
        for i in range(n_samples):
            samples[i] = self.trajectory_from_weights(T, weight_samples[i])
        return samples

    def from_weight_distribution(self, mean, cov):
        """Initialize ProMP from mean and covariance in weight space.

        Parameters
        ----------
        mean : array, shape (n_dims * n_weights_per_dim)
            Mean of weight distribution

        cov : array, shape (n_dims * n_weights_per_dim, n_dims * n_weights_per_dim)
            Covariance of weight distribution

        Returns
        -------
        self : ProMP
            This object
        """
        self.weight_mean = mean
        self.weight_cov = cov
        return self

    def imitate(self, Ts, Ys, n_iter=1000, min_delta=1e-5, verbose=0):
        r"""Learn ProMP from multiple demonstrations.

        For details, see Section 3.2 of [1]_. We use the parameters
        :math:`P = I` (identity matrix), :math:`\mu_0 = 0, k_0 = 0, \nu_0 = 0,
        \Sigma_0 = 0,\alpha_0 = 0,\beta_0 = 0`.

        Parameters
        ----------
        Ts : array, shape (n_demos, n_steps)
            Time steps of demonstrations

        Ys : array, shape (n_demos, n_steps, n_dims)
            Demonstrations

        n_iter : int, optional (default: 1000)
            Maximum number of iterations

        min_delta : float, optional (default: 1e-5)
            Minimum delta between means to continue iteration

        verbose : int, optional (default: 0)
            Verbosity level

        References
        ----------
        .. [1] Lazaric, A., Ghavamzadeh, M. (2010).
           Bayesian Multi-Task Reinforcement Learning. In Proceedings of the
           27th International Conference on International Conference on Machine
           Learning (ICML'10) (pp. 599-606).
           https://hal.inria.fr/inria-00475214/document
        """
        gamma = 0.7

        n_demos = len(Ts)
        self.variance = 1.0

        means = np.zeros((n_demos, self.n_weights))
        covs = np.empty((n_demos, self.n_weights, self.n_weights))

        # Precompute constant terms in expectation-maximization algorithm

        # n_demos x n_steps*self.n_dims x n_steps*self.n_dims
        Hs = []
        for demo_idx in range(n_demos):
            n_steps = len(Ys[demo_idx])
            H_partial = np.eye(n_steps)
            for y in range(n_steps - 1):
                H_partial[y, y + 1] = -gamma
            H = _nd_block_diagonal(H_partial, self.n_dims)
            Hs.append(H)

        # n_demos x n_steps*n_dims
        Ys_rearranged = [Y.T.ravel() for Y in Ys]

        # n_demos x n_steps*self.n_dims
        Rs = []
        for demo_idx in range(n_demos):
            R = Hs[demo_idx].dot(Ys_rearranged[demo_idx])
            Rs.append(R)

        # n_demos
        # RR in original code
        RTRs = []
        for demo_idx in range(n_demos):
            RTR = Rs[demo_idx].T.dot(Rs[demo_idx])
            RTRs.append(RTR)

        # n_demos x self.n_dims*self.n_weights_per_dim
        # x self.n_dims*self.n_steps
        # BH in original code
        PhiHTs = []
        for demo_idx in range(n_demos):
            PhiHT = self._rbfs_nd_sequence(Ts[demo_idx]).dot(Hs[demo_idx].T)
            PhiHTs.append(PhiHT)

        # n_demos x self.n_dims*self.n_weights_per_dim
        # mean_esteps in original code
        PhiHTRs = []
        for demo_idx in range(n_demos):
            PhiHTR = PhiHTs[demo_idx].dot(Rs[demo_idx])
            PhiHTRs.append(PhiHTR)

        # n_demos x self.n_dims*self.n_weights_per_dim
        # x self.n_dims*self.n_weights_per_dim
        # cov_esteps in original code
        PhiHTHPhiTs = []
        for demo_idx in range(n_demos):
            PhiHTHPhiT = PhiHTs[demo_idx].dot(PhiHTs[demo_idx].T)
            PhiHTHPhiTs.append(PhiHTHPhiT)

        n_samples = sum([Y.shape[0] for Y in Ys])

        for it in range(n_iter):
            weight_mean_old = self.weight_mean

            for demo_idx in range(n_demos):
                means[demo_idx], covs[demo_idx] = self._expectation(
                        PhiHTRs[demo_idx], PhiHTHPhiTs[demo_idx])

            self._maximization(
                means, covs, RTRs, PhiHTRs, PhiHTHPhiTs, n_samples)

            delta = np.linalg.norm(self.weight_mean - weight_mean_old)
            if verbose:
                print("Iteration %04d: delta = %g" % (it + 1, delta))
            if delta < min_delta:
                break

    def _rbfs_1d_point(self, t, t_max=1.0, overlap=0.7):
        """Radial basis functions for one dimension and a point.

        Parameters
        ----------
        t : float
            Time at which the activations of RBFs will be queried. Note that
            we internally normalize the time so that t_max == 1.

        t_max : float, optional (default: 1)
            Duration of the ProMP

        overlap : float, optional (default: 0.7)
            Indicates how much the RBFs are allowed to overlap.

        Returns
        -------
        activations : array, shape (n_weights_per_dim,)
            Activations of RBFs for each time step.
        """
        h = -1.0 / (8.0 * self.n_weights_per_dim ** 2 * np.log(overlap))

        # normalize time to interval [0, 1]
        t = t / t_max

        activations = np.exp(-(t - self.centers[:]) ** 2 / (2.0 * h))
        activations /= activations.sum(axis=0)  # normalize activations for each step

        assert activations.ndim == 1
        assert activations.shape[0] == self.n_weights_per_dim

        return activations

    def _rbfs_nd_sequence(self, T, overlap=0.7):
        """Radial basis functions for n_dims dimensions and a sequence.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Times at which the activations of RBFs will be queried. Note that
            we assume that T[0] == 0.0 and the times will be normalized
            internally so that T[-1] == 1.0.

        overlap : float, optional (default: 0.7)
            Indicates how much the RBFs are allowed to overlap.

        Returns
        -------
        activations : array, shape (n_dims * n_weights_per_dim, n_dims * n_steps)
            Activations of RBFs for each time step and each dimension.
        """
        return _nd_block_diagonal(
            self._rbfs_1d_sequence(T, overlap), self.n_dims)

    def _rbfs_1d_sequence(self, T, overlap=0.7, normalize=True):
        """Radial basis functions for one dimension and a sequence.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Times at which the activations of RBFs will be queried. Note that
            we assume that T[0] == 0.0 and the times will be normalized
            internally so that T[-1] == 1.0.

        overlap : float, optional (default: 0.7)
            Indicates how much the RBFs are allowed to overlap.

        normalize : bool, optional (default: True)
            Normalize activations to sum up to one in each step

        Returns
        -------
        activations : array, shape (n_weights_per_dim, n_steps)
            Activations of RBFs for each time step.
        """
        assert T.ndim == 1

        n_steps = len(T)

        h = -1.0 / (8.0 * self.n_weights_per_dim ** 2 * np.log(overlap))

        # normalize time to interval [0, 1]
        T = np.atleast_2d(T)
        T /= np.max(T)

        activations = np.exp(
            -(T - self.centers[:, np.newaxis]) ** 2 / (2.0 * h))
        if normalize:
            activations /= activations.sum(axis=0)

        assert activations.shape[0] == self.n_weights_per_dim
        assert activations.shape[1] == n_steps

        return activations

    def _rbfs_derivative_nd_sequence(self, T, overlap=0.7):
        """Derivative of RBFs for n_dims dimensions and a sequence.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Times at which the activations of RBFs will be queried. Note that
            we assume that T[0] == 0.0 and the times will be normalized
            internally so that T[-1] == 1.0.

        overlap : float, optional (default: 0.7)
            Indicates how much the RBFs are allowed to overlap.

        Returns
        -------
        activations : array, shape (n_dims * n_weights_per_dim, n_dims * n_steps)
            Activations of derivative of RBFs for each time step and dimension.
        """
        return _nd_block_diagonal(
            self._rbfs_derivative_1d_sequence(T, overlap), self.n_dims)

    def _rbfs_derivative_1d_sequence(self, T, overlap=0.7):
        """Derivative of RBFs for one dimension and a sequence.

        Parameters
        ----------
        T : array-like, shape (n_steps,)
            Times at which the activations of RBFs will be queried. Note that
            we assume that T[0] == 0.0 and the times will be normalized
            internally so that T[-1] == 1.0.

        overlap : float, optional (default: 0.7)
            Indicates how much the RBFs are allowed to overlap.

        Returns
        -------
        activations : array, shape (n_weights_per_dim, n_steps)
            Activations of derivative of RBFs for each time step.
        """
        assert T.ndim == 1

        n_steps = len(T)

        h = -1.0 / (8.0 * self.n_weights_per_dim ** 2 * np.log(overlap))

        rbfs = self._rbfs_1d_sequence(T, overlap, normalize=False)
        rbfs_sum_per_step = rbfs.sum(axis=0)

        # normalize time to interval [0, 1]
        T = np.atleast_2d(T)
        T /= np.max(T)

        rbfs_deriv = (self.centers[:, np.newaxis] - T) / h
        rbfs_deriv *= rbfs
        rbfs_deriv_sum_per_step = rbfs_deriv.sum(axis=0)
        rbfs_deriv = (
             rbfs_deriv * rbfs_sum_per_step
             - rbfs * rbfs_deriv_sum_per_step) / (rbfs_sum_per_step ** 2)

        assert rbfs_deriv.shape[0] == self.n_weights_per_dim
        assert rbfs_deriv.shape[1] == n_steps

        return rbfs_deriv

    def _expectation(self, PhiHTR, PhiHTHPhiT):
        cov = np.linalg.pinv(PhiHTHPhiT / self.variance
                             + np.linalg.pinv(self.weight_cov))
        mean = cov.dot(PhiHTR / self.variance
                       + np.linalg.pinv(self.weight_cov).dot(self.weight_mean))
        return mean, cov

    def _maximization(self, means, covs, RRs, PhiHTR, PhiHTHPhiTs, n_samples):
        M = len(means)

        self.weight_mean = np.mean(means, axis=0)

        centered = means - self.weight_mean
        self.weight_cov = (centered.T.dot(centered) + np.sum(covs, axis=0)) / M
        # TODO what is d + 2?

        self.variance = 0.0
        for i in range(len(means)):
            # trace is the same, independent of the order of matrix
            # multiplications, see:
            # https://www.math.uwaterloo.ca/~hwolkowi/matrixcookbook.pdf,
            # Equation 16
            self.variance += np.trace(PhiHTHPhiTs[i].dot(covs[i]))

            self.variance += RRs[i]
            self.variance -= 2.0 * PhiHTR[i].T.dot(means[i].T)
            self.variance += (means[i].dot(PhiHTHPhiTs[i].dot(means[i].T)))

        # TODO why these factors?
        self.variance /= (np.linalg.norm(means) * M * self.n_dims * n_samples
                          + 2.0)
        #self.variance /= self.n_dims * n_samples + 2.0


def _nd_block_diagonal(partial_1d, n_dims):
    """Replicates matrix n_dims times to form a block-diagonal matrix.

    We also accept matrices of rectangular shape. In this case the result is
    not officially called a block-diagonal matrix anymore.

    Parameters
    ----------
    partial_1d : array, shape (n_block_rows, n_block_cols)
        Matrix that should be replicated.

    n_dims : int
        Number of times that the matrix has to be replicated.

    Returns
    -------
    full_nd : array, shape (n_block_rows * n_dims, n_block_cols * n_dims)
        Block-diagonal matrix with n_dims replications of the initial matrix.
    """
    assert partial_1d.ndim == 2
    n_block_rows, n_block_cols = partial_1d.shape

    full_nd = np.zeros((n_block_rows * n_dims, n_block_cols * n_dims))
    for j in range(n_dims):
        full_nd[n_block_rows * j:n_block_rows * (j + 1),
                n_block_cols * j:n_block_cols * (j + 1)] = partial_1d
    return full_nd


def via_points(promp, ts, y_cond, y_conditional_cov=None):
    """Condition ProMP on several via-points.

    For details, see section 2.2 on page 4 of [1]_

    Parameters
    ----------
    promp : ProMP instance

    ts : array, shape (n_positions,)
        Time vector at which the activations of RBFs will be queried. Note that
        we internally normalize the time so that t_max == 1.

    y_cond : array, shape (n_positions,)
        Desired mean position vector corresponding to each time in `ts`.

    y_conditional_cov : array, shape (n_dims, n_dims), optional (default: 0)
        Covariances of desired positions.

    Returns
    -------
    conditional_promp : ProMP
        New conditional ProMP

    References
    ----------
    .. [1] Paraschos, A., Daniel, C., Peters, J., Neumann, G. (2013).
       Probabilistic movement primitives, In C.J. Burges and L. Bottou and
       M. Welling and Z. Ghahramani and K.Q. Weinberger (Eds.), Advances in
       Neural Information Processing Systems, 26,
       https://papers.nips.cc/paper/2013/file/e53a0a2978c28872a4505bdb51db06dc-Paper.pdf
    """
    if y_conditional_cov is None:
        y_conditional_cov = np.zeros(y_cond.shape)

    for idx, t_i in enumerate(ts):
        promp = promp.condition_position(
            y_mean=y_cond[idx],
            y_cov=y_conditional_cov[[idx]],
            t=t_i,
            t_max=1.0
        )
    return promp



r"""Dynamical Movement Primitive (DMP)
==================================

This module provides implementations of various DMP types. DMPs consist of
a goal-directed movement generated by the transformation system and a forcing
term that defines the shape of the trajectory. They are time-dependent and
usually converge to the goal after a constant execution time.

Every implementation is slightly different, but we use a similar notation for
all of them:

* :math:`y,\dot{y},\ddot{y}` - position, velocity, and acceleration; these
  might be translation components only, orientation, or both
* :math:`y_0` - start position
* :math:`g` - goal of the movement (attractor of the DMP)
* :math:`\tau` - execution time
* :math:`z` - phase variable, starts at 1 and converges to 0
* :math:`f(z)` - forcing term, learned component that defines the shape of the
  movement
* :math:`C_t` - coupling term that is added to the acceleration of the DMP
"""
from ._base import DMPBase, WeightParametersMixin
from ._dmp import DMP, dmp_transformation_system
from ._dmp_with_final_velocity import DMPWithFinalVelocity
from ._cartesian_dmp import CartesianDMP
from ._dual_cartesian_dmp import DualCartesianDMP
from ._coupling_terms import (
    CouplingTermObstacleAvoidance2D, CouplingTermObstacleAvoidance3D,
    CouplingTermPos1DToPos1D, CouplingTermPos3DToPos3D,
    CouplingTermDualCartesianPose, CouplingTermDualCartesianDistance,
    CouplingTermDualCartesianTrajectory, obstacle_avoidance_acceleration_2d,
    obstacle_avoidance_acceleration_3d)
from ._state_following_dmp import StateFollowingDMP
from ._canonical_system import canonical_system_alpha, phase


__all__ = [
    "DMPBase", "WeightParametersMixin",
    "DMP", "dmp_transformation_system", "DMPWithFinalVelocity", "CartesianDMP",
    "DualCartesianDMP", "CouplingTermPos1DToPos1D",
    "CouplingTermObstacleAvoidance2D", "CouplingTermObstacleAvoidance3D",
    "CouplingTermPos3DToPos3D", "CouplingTermDualCartesianPose",
    "CouplingTermDualCartesianDistance", "CouplingTermDualCartesianTrajectory",
    "canonical_system_alpha", "phase", "obstacle_avoidance_acceleration_2d",
    "obstacle_avoidance_acceleration_3d"]



"""Cython implementations of basic DMP functions."""
import numpy as np
cimport numpy as np
cimport cython
from libcpp cimport bool
from libc.math cimport sqrt, cos, sin, acos, pi, exp


np.import_array()


cdef double M_2PI = 2.0 * pi
cdef double M_PI_HALF = 0.5 * pi
cdef double EPSILON = 1e-10


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef phase(t, double alpha, double goal_t, double start_t):
    """Map time to phase.

    Parameters
    ----------
    t : float
        Current time.

    alpha : float
        Value of the alpha parameter of the canonical system.

    goal_t : float
        Time at which the execution should be done.

    start_t : float
        Time at which the execution should start.

    Returns
    -------
    z : float
        Value of phase variable.
    """
    cdef double execution_time = goal_t - start_t
    return np.exp(-alpha * (t - start_t) / execution_time)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef dmp_step(
        double last_t, double t, np.ndarray[double, ndim=1] current_y, np.ndarray[double, ndim=1] current_yd,
        np.ndarray[double, ndim=1] goal_y, np.ndarray[double, ndim=1] goal_yd,
        np.ndarray[double, ndim=1] goal_ydd, np.ndarray[double, ndim=1] start_y,
        np.ndarray[double, ndim=1] start_yd, np.ndarray[double, ndim=1] start_ydd,
        double goal_t, double start_t, np.ndarray[double, ndim=1] alpha_y, np.ndarray[double, ndim=1] beta_y,
        object forcing_term, object coupling_term=None,
        tuple coupling_term_precomputed=None,
        double int_dt=0.001, double p_gain=0.0,
        np.ndarray tracking_error=None,
        bint smooth_scaling=False):
    """Integrate regular DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (n_dims,)
        Current position. Will be modified.

    current_yd : array, shape (n_dims,)
        Current velocity. Will be modified.

    goal_y : array, shape (n_dims,)
        Goal position.

    goal_yd : array, shape (n_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_dims,)
        Goal acceleration.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    if t <= start_t:
        current_y[:] = start_y
        current_yd[:] = start_yd

    cdef double execution_time = goal_t - start_t

    cdef int n_dims = current_y.shape[0]

    cdef np.ndarray[double, ndim=1] current_ydd = np.empty(n_dims, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] cd = np.empty(n_dims, dtype=np.float64)
    cdef np.ndarray[double, ndim=1] cdd = np.empty(n_dims, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] f = np.empty(n_dims, dtype=np.float64)

    cdef int d
    cdef double current_t
    cdef double dt
    cdef double z
    cdef double smoothing

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd[:], cdd[:] = coupling_term.coupling(current_y, current_yd)
        elif coupling_term_precomputed is not None:
            cd[:] = coupling_term_precomputed[0]
            cdd[:] = coupling_term_precomputed[1]
        else:
            cd[:] = 0.0
            cdd[:] = 0.0
        if tracking_error is not None:
            cdd += p_gain * tracking_error / dt

        z = forcing_term.phase(current_t)
        f[:] = forcing_term.forcing_term(z).squeeze()

        for d in range(n_dims):
            if smooth_scaling:
                smoothing = beta_y[d] * (goal_y[d] - start_y[d]) * z
            else:
                smoothing = 0.0

            current_ydd[d] = (
                alpha_y[d] * (
                    beta_y[d] * (goal_y[d] - current_y[d])
                    - execution_time * current_yd[d]
                    - smoothing
                )
                + f[d]
                + cdd[d]
            ) / execution_time ** 2
            current_yd[d] += dt * current_ydd[d] + cd[d] / execution_time
            current_y[d] += dt * current_yd[d]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef dmp_step_rk4(
        double last_t, double t, np.ndarray[double, ndim=1] current_y, np.ndarray[double, ndim=1] current_yd,
        np.ndarray[double, ndim=1] goal_y, np.ndarray[double, ndim=1] goal_yd,
        np.ndarray[double, ndim=1] goal_ydd, np.ndarray[double, ndim=1] start_y,
        np.ndarray[double, ndim=1] start_yd, np.ndarray[double, ndim=1] start_ydd,
        double goal_t, double start_t, np.ndarray[double, ndim=1] alpha_y, np.ndarray[double, ndim=1] beta_y,
        object forcing_term, object coupling_term=None,
        tuple coupling_term_precomputed=None,
        double int_dt=0.001, double p_gain=0.0,
        np.ndarray tracking_error=None,
        bint smooth_scaling=False):
    """Integrate regular DMP for one step with RK4 integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (n_dims,)
        Current position. Will be modified.

    current_yd : array, shape (n_dims,)
        Current velocity. Will be modified.

    goal_y : array, shape (n_dims,)
        Goal position.

    goal_yd : array, shape (n_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_dims,)
        Goal acceleration.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    cdef int n_dims = current_y.shape[0]

    cdef np.ndarray[double, ndim=1] cd = np.zeros(n_dims, dtype=np.float64)
    cdef np.ndarray[double, ndim=1] cdd = np.zeros(n_dims, dtype=np.float64)
    if coupling_term_precomputed is not None:
        cd += coupling_term_precomputed[0]
        cdd += coupling_term_precomputed[1]

    # precompute constants for following queries
    cdef double execution_time = goal_t - start_t
    cdef double dt = t - last_t
    cdef double dt_2 = 0.5 * dt
    cdef np.ndarray[double, ndim=1] T = np.array([t, t + dt_2, t + dt])
    cdef np.ndarray[double, ndim=1] Z = forcing_term.phase(T)
    cdef np.ndarray[double, ndim=2] F = forcing_term.forcing_term(Z)
    cdef np.ndarray[double, ndim=1] tdd
    if tracking_error is not None:
        tdd = p_gain / dt * tracking_error
    else:
        tdd = np.zeros(n_dims, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] K0 = _dmp_acc(
        current_y, current_yd, cdd, dt, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[0], execution_time, F[:, 0], coupling_term,
        p_gain, tdd, smooth_scaling)
    cdef np.ndarray[double, ndim=1] C1 = current_yd + dt_2 * K0
    cdef np.ndarray[double, ndim=1] K1 = _dmp_acc(
        current_y + dt_2 * current_yd, C1, cdd, dt, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[1], execution_time, F[:, 1], coupling_term,
        p_gain, tdd, smooth_scaling)
    cdef np.ndarray[double, ndim=1] C2 = current_yd + dt_2 * K1
    cdef np.ndarray[double, ndim=1] K2 = _dmp_acc(
        current_y + dt_2 * C1, C2, cdd, dt, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[1], execution_time, F[:, 1], coupling_term,
        p_gain, tdd, smooth_scaling)
    cdef np.ndarray[double, ndim=1] C3 = current_yd + dt * K2
    cdef np.ndarray[double, ndim=1] K3 = _dmp_acc(
        current_y + dt * C2, C3, cdd, dt, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[2], execution_time, F[:, 2], coupling_term,
        p_gain, tdd, smooth_scaling)

    cdef int i
    for i in range(n_dims):
        current_y[i] += dt * (current_yd[i] + 2 * C1[i] + 2 * C2[i] + C3[i]) / 6.0
        current_yd[i] += dt * (K0[i] + 2 * K1[i] + 2 * K2[i] + K3[i]) / 6.0

    if coupling_term is not None:
        cd[:], _ = coupling_term.coupling(current_y, current_yd)
        current_yd += cd / execution_time


cdef _dmp_acc(
        np.ndarray[double, ndim=1] Y, np.ndarray[double, ndim=1] V,
        np.ndarray[double, ndim=1] cdd, double dt, np.ndarray[double, ndim=1] alpha_y, np.ndarray[double, ndim=1] beta_y,
        np.ndarray[double, ndim=1] goal_y, np.ndarray[double, ndim=1] goal_yd,
        np.ndarray[double, ndim=1] goal_ydd, np.ndarray[double, ndim=1] start_y,
        double z, double execution_time, np.ndarray[double, ndim=1] f,
        object coupling_term, double p_gain, np.ndarray[double, ndim=1] tdd,
        bint smooth_scaling):
    if coupling_term is not None:
        _, cdd[:] = coupling_term.coupling(Y, V)

    cdef int n_dims = Y.shape[0]
    cdef np.ndarray[double, ndim=1] Ydd = np.empty(n_dims, dtype=np.float64)
    cdef int d
    cdef double smoothing
    for d in range(n_dims):
        if smooth_scaling:
            smoothing = beta_y[d] * (goal_y[d] - start_y[d]) * z
        else:
            smoothing = 0.0
        Ydd[d] = (
            alpha_y[d] * (
                beta_y[d] * (goal_y[d] - Y[d])
                - execution_time * V[d]
                - smoothing
            )
            + f[d]
            + cdd[d]
            + tdd[d]
        ) / execution_time ** 2
    return Ydd


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef dmp_step_quaternion(
        double last_t, double t,
        np.ndarray[double, ndim=1] current_y,
        np.ndarray[double, ndim=1] current_yd,
        np.ndarray[double, ndim=1] goal_y,
        np.ndarray[double, ndim=1] goal_yd,
        np.ndarray[double, ndim=1] goal_ydd,
        np.ndarray[double, ndim=1] start_y,
        np.ndarray[double, ndim=1] start_yd,
        np.ndarray[double, ndim=1] start_ydd,
        double goal_t, double start_t, np.ndarray[double, ndim=1] alpha_y, np.ndarray[double, ndim=1] beta_y,
        forcing_term, coupling_term=None, coupling_term_precomputed=None,
        double int_dt=0.001, bint smooth_scaling=False):
    """Integrate quaternion DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (7,)
        Current position. Will be modified.

    current_yd : array, shape (6,)
        Current velocity. Will be modified.

    goal_y : array, shape (7,)
        Goal position.

    goal_yd : array, shape (6,)
        Goal velocity.

    goal_ydd : array, shape (6,)
        Goal acceleration.

    start_y : array, shape (7,)
        Start position.

    start_yd : array, shape (6,)
        Start velocity.

    start_ydd : array, shape (6,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (6,)
        Constant in transformation system.

    beta_y : array, shape (6,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if t <= start_t:
        current_y[:] = start_y
        current_yd[:] = start_yd

    cdef double execution_time = goal_t - start_t

    cdef np.ndarray[double, ndim=1] current_ydd = np.empty(3, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] cd = np.zeros(3, dtype=np.float64)
    cdef np.ndarray[double, ndim=1] cdd = np.zeros(3, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] f = np.empty(3, dtype=np.float64)

    cdef np.ndarray[double, ndim=1] goal_y_minus_start_y

    cdef int d
    cdef double current_t
    cdef double dt
    cdef double z
    cdef np.ndarray[double, ndim=1] smoothing = np.empty(3, dtype=np.float64)
    if not smooth_scaling:
        smoothing[:] = 0

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd[:], cdd[:] = coupling_term.coupling(current_y, current_yd)
        elif coupling_term_precomputed is not None:
            cd[:] = coupling_term_precomputed[0]
            cdd[:] = coupling_term_precomputed[1]

        z = forcing_term.phase(current_t)
        f[:] = forcing_term.forcing_term(z).squeeze()

        if smooth_scaling:
            goal_y_minus_start_y = compact_axis_angle_from_quaternion(
                concatenate_quaternions(goal_y, q_conj(start_y)))
            smoothing[:] = beta_y * z * goal_y_minus_start_y

        current_ydd[:] = (
            alpha_y * (
                beta_y * compact_axis_angle_from_quaternion(
                    concatenate_quaternions(goal_y, q_conj(current_y)))
                - execution_time * current_yd
                - smoothing
            )
            + f
            + cdd
        ) / execution_time ** 2
        current_yd += dt * current_ydd + cd / execution_time
        current_y[:] = concatenate_quaternions(
            quaternion_from_compact_axis_angle(dt * current_yd), current_y)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef dmp_step_dual_cartesian(
        double last_t, double t,
        np.ndarray[double, ndim=1] current_y, np.ndarray[double, ndim=1] current_yd,
        np.ndarray[double, ndim=1] goal_y, np.ndarray[double, ndim=1] goal_yd, np.ndarray[double, ndim=1] goal_ydd,
        np.ndarray[double, ndim=1] start_y, np.ndarray[double, ndim=1] start_yd, np.ndarray[double, ndim=1] start_ydd,
        double goal_t, double start_t, np.ndarray[double, ndim=1] alpha_y, np.ndarray[double, ndim=1] beta_y,
        forcing_term, coupling_term=None,
        double int_dt=0.001,
        double p_gain=0.0, np.ndarray tracking_error=None,
        bint smooth_scaling=False):
    """Integrate bimanual Cartesian DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (14,)
        Current position. Will be modified.

    current_yd : array, shape (12,)
        Current velocity. Will be modified.

    goal_y : array, shape (14,)
        Goal position.

    goal_yd : array, shape (12,)
        Goal velocity.

    goal_ydd : array, shape (12,)
        Goal acceleration.

    start_y : array, shape (14,)
        Start position.

    start_yd : array, shape (12,)
        Start velocity.

    start_ydd : array, shape (12,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (12,)
        Constant in transformation system.

    beta_y : array, shape (12,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if t <= start_t:
        current_y[:] = start_y
        current_yd[:] = start_yd

    cdef double execution_time = goal_t - start_t

    cdef np.ndarray[double, ndim=1] current_ydd = np.empty_like(current_yd)

    cdef int n_vel_dims = current_yd.shape[0]

    cdef np.ndarray[double, ndim=1] cd = np.empty(n_vel_dims, dtype=float)
    cdef np.ndarray[double, ndim=1] cdd = np.empty(n_vel_dims, dtype=float)

    cdef np.ndarray[double, ndim=1] f = np.empty(n_vel_dims, dtype=float)
    cdef np.ndarray[double, ndim=1] goal_y_minus_start_y
    cdef double smoothing_pos
    cdef np.ndarray[double, ndim=1] smoothing_orn = np.empty(3, dtype=float)
    if not smooth_scaling:
        smoothing_orn[:] = 0.0

    cdef double z

    cdef int pps
    cdef int pvs
    cdef np.ndarray[long, ndim=2] POS_INDICES = np.array(
        [[0, 0], [1, 1], [2, 2], [7, 6], [8, 7], [9, 8]], dtype=int)

    cdef double dt
    cdef double current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is None:
            cd[:] = 0.0
            cdd[:] = 0.0
        else:
            cd[:], cdd[:] = coupling_term.coupling(current_y, current_yd)

        z = forcing_term.phase(current_t)
        f[:] = forcing_term.forcing_term(z).squeeze()
        if tracking_error is not None:
            for pps, pvs in POS_INDICES:
                cdd[pvs] += p_gain * tracking_error[pps] / dt
            for ops, ovs in ((slice(3, 7), slice(3, 6)), (slice(10, 14), slice(9, 12))):
                cdd[ovs] += p_gain * compact_axis_angle_from_quaternion(tracking_error[ops]) / dt

        # position components
        for pps, pvs in POS_INDICES:
            if smooth_scaling:
                smoothing_pos = beta_y[pps] * (goal_y[pps] - start_y[pps]) * z
            else:
                smoothing_pos = 0.0
            current_ydd[pvs] = (
                alpha_y[pvs] * (
                    beta_y[pvs] * (goal_y[pps] - current_y[pps])
                    - execution_time * current_yd[pvs]
                    - smoothing_pos
                )
                + f[pvs]
                + cdd[pvs]
            ) / execution_time ** 2
            current_yd[pvs] += dt * current_ydd[pvs] + cd[pvs] / execution_time
            current_y[pps] += dt * current_yd[pvs]

        # orientation components
        for ops, ovs in ((slice(3, 7), slice(3, 6)), (slice(10, 14), slice(9, 12))):
            if smooth_scaling:
                goal_y_minus_start_y = compact_axis_angle_from_quaternion(
                    concatenate_quaternions(goal_y[ops], q_conj(start_y[ops])))
                smoothing_orn[:] = beta_y[ovs] * z * goal_y_minus_start_y
            current_ydd[ovs] = (
                alpha_y[ovs] * (
                    beta_y[ovs] * compact_axis_angle_from_quaternion(
                        concatenate_quaternions(goal_y[ops], q_conj(current_y[ops])))
                    - execution_time * current_yd[ovs]
                    - smoothing_orn
                )
                + f[ovs]
                + cdd[ovs]
            ) / execution_time ** 2
            current_yd[ovs] += dt * current_ydd[ovs] + cd[ovs] / execution_time
            current_y[ops] = concatenate_quaternions(
                quaternion_from_compact_axis_angle(dt * current_yd[ovs]),
                current_y[ops])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef concatenate_quaternions(
        np.ndarray[double, ndim=1] q1, np.ndarray[double, ndim=1] q2):
    """Concatenate two quaternions.

    We use Hamilton's quaternion multiplication.

    Parameters
    ----------
    q1 : array-like, shape (4,)
        First quaternion

    q2 : array-like, shape (4,)
        Second quaternion

    Returns
    -------
    q12 : array-like, shape (4,)
        Quaternion that represents the concatenated rotation q1 * q2
    """
    cdef np.ndarray[double, ndim=1] q12 = np.empty(4)
    q12[0] = q1[0] * q2[0]
    # cross product q1[1:] x q2[1:]
    q12[1] = q1[2] * q2[3] - q1[3] * q2[2]
    q12[2] = q1[3] * q2[1] - q1[1] * q2[3]
    q12[3] = q1[1] * q2[2] - q1[2] * q2[1]
    cdef int i
    for i in range(1, 4):
        q12[0] -= q1[i] * q2[i]
        q12[i] += q1[0] * q2[i] + q2[0] * q1[i]

    cdef double norm = sqrt(q12[0] * q12[0] + q12[1] * q12[1] + q12[2] * q12[2] + q12[3] * q12[3])
    for i in range(4):
        q12[i] /= norm
    return q12


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef quaternion_from_compact_axis_angle(np.ndarray[double, ndim=1] a):
    """Compute quaternion from compact axis-angle (exponential map).

    We usually assume active rotations.

    Parameters
    ----------
    a : array-like, shape (4,)
        Axis of rotation and rotation angle: angle * (x, y, z)

    Returns
    -------
    q : array-like, shape (4,)
        Unit quaternion to represent rotation: (w, x, y, z)
    """
    cdef double angle = sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])
    if angle == 0.0:
        return np.array([1.0, 0.0, 0.0, 0.0])

    cdef np.ndarray[double, ndim=1] axis
    axis = a / angle

    cdef np.ndarray[double, ndim=1] q = np.empty(4)
    cdef double half_angle = angle / 2.0
    q[0] = cos(half_angle)
    cdef double s = sin(half_angle)
    q[1] = s * axis[0]
    q[2] = s * axis[1]
    q[3] = s * axis[2]
    return q


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef q_conj(np.ndarray[double, ndim=1] q):
    """Conjugate of quaternion.

    The conjugate of a unit quaternion inverts the rotation represented by
    this unit quaternion. The conjugate of a quaternion q is often denoted
    as q*.

    Parameters
    ----------
    q : array-like, shape (4,)
        Unit quaternion to represent rotation: (w, x, y, z)

    Returns
    -------
    q_c : array-like, shape (4,)
        Conjugate (w, -x, -y, -z)
    """
    return np.array([q[0], -q[1], -q[2], -q[3]])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef compact_axis_angle_from_quaternion(np.ndarray[double, ndim=1] q):
    """Compute compact axis-angle from quaternion (logarithmic map).

    We usually assume active rotations.

    Parameters
    ----------
    q : array-like, shape (4,)
        Unit quaternion to represent rotation: (w, x, y, z)

    Returns
    -------
    a : array-like, shape (3,)
        Axis of rotation and rotation angle: angle * (x, y, z). The angle is
        constrained to [0, pi).
    """
    cdef double p_norm_sqr = q[1] * q[1] + q[2] * q[2] + q[3] * q[3]
    if p_norm_sqr < 1e-32:
        return np.zeros(3)
    cdef double q_norm = sqrt(q[0] * q[0] + p_norm_sqr)
    cdef np.ndarray[double, ndim=1] q_n = q / q_norm
    cdef double p_norm = sqrt(p_norm_sqr) / q_norm
    # Source of the solution: http://stackoverflow.com/a/32266181
    cdef double angle = ((2 * acos(q_n[0]) + pi) % M_2PI - pi)
    return q_n[1:] / (p_norm / angle)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef matrix_from_compact_axis_angle(np.ndarray[double, ndim=1] r):
    cdef double theta = np.linalg.norm(r)
    if theta == 0.0:
        return np.eye(3)
    cdef double ux = r[0] / theta
    cdef double uy = r[1] / theta
    cdef double uz = r[2] / theta
    cdef double c = cos(theta)
    cdef double s = sin(theta)
    cdef double ci = 1.0 - c
    cdef np.ndarray[double, ndim=2] R = np.empty((3, 3), dtype=float)
    R[0, 0] = ci * ux * ux + c
    R[0, 1] = ci * ux * uy - uz * s
    R[0, 2] = ci * ux * uz + uy * s
    R[1, 0] = ci * uy * ux + uz * s
    R[1, 1] = ci * uy * uy + c
    R[1, 2] = ci * uy * uz - ux * s
    R[2, 0] = ci * uz * ux - uy * s
    R[2, 1] = ci * uz * uy + ux * s
    R[2, 2] = ci * uz * uz + c
    return R


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef obstacle_avoidance_acceleration_2d(
        np.ndarray[double, ndim=1] y, np.ndarray[double, ndim=1] yd,
        np.ndarray[double, ndim=1] obstacle_position, double gamma, double beta):
    """Compute acceleration for obstacle avoidance in 2D.

    Parameters
    ----------
    y : array, shape (2,)
        Current position.

    yd : array, shape (2,)
        Current velocity.

    obstacle_position : array, shape (2,)
        Position of the point obstacle.

    gamma : float
        Obstacle avoidance parameter.

    beta : float
        Obstacle avoidance parameter.

    Returns
    -------
    cdd : array, shape (2,)
        Acceleration.
    """
    cdef np.ndarray[double, ndim=1] obstacle_diff = obstacle_position - y
    cdef double r = obstacle_diff[0] * yd[1] - obstacle_diff[1] * yd[0]
    if r != 0.0:
        r *= M_PI_HALF / abs(r)
    cdef np.ndarray[double, ndim=1] r_vec = np.array([0.0, 0.0, r])
    cdef np.ndarray[double, ndim=2] R = matrix_from_compact_axis_angle(r_vec)[:2, :2]
    cdef double theta_nom = obstacle_diff[0] * yd[0] + obstacle_diff[1] * yd[1]
    cdef double obstacle_diff_norm = sqrt(obstacle_diff[0] * obstacle_diff[0] + obstacle_diff[1] * obstacle_diff[1])
    cdef double yd_norm = sqrt(yd[0] * yd[0] + yd[1] * yd[1])
    cdef double theta_denom = obstacle_diff_norm * yd_norm + EPSILON
    cdef double theta = acos(theta_nom / theta_denom)
    cdef np.ndarray[double, ndim=1] rotated_velocity = np.dot(R, yd)
    cdef np.ndarray[double, ndim=1] cdd = gamma * rotated_velocity * (theta * exp(-beta * theta))
    return cdd



import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DualCartesianDMP, CouplingTermDualCartesianPose
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
from movement_primitives.testing.simulation import RH5Simulation, draw_pose

dt = 0.0005
execution_time = 1.0
plot = False

#ct = CouplingTermDualCartesianDistance(desired_distance=0.2, lf=(1.0, 0.0), k=1, c1=0.1, c2=1000)
#ct = CouplingTermDualCartesianOrientation(desired_distance=np.deg2rad(25), lf=(1.0, 0.0), k=0.1)
desired_distance = pt.transform_from(  # right arm to left arm
    R=pr.active_matrix_from_intrinsic_euler_zyx([np.pi, 0, np.deg2rad(15)]),
    p=np.array([0.0, -0.3, 0.0])
)
ct = CouplingTermDualCartesianPose(desired_distance=desired_distance, lf=(1.0, 0.0), k=1, c1=0.1, c2=10000)

rh5 = RH5Simulation(dt=dt, gui=True, real_time=False)
q0 = np.array([-1.57, 0.76, 0, -1.3, 0, 0, -0.55, 1.57, -0.76, 0, 1.3, 0, 0, -0.55])
rh5.set_desired_joint_state(q0, position_control=True)
rh5.sim_loop(int(1.0 / dt))

ee_state = rh5.get_ee_state()
left2base_start = ee_state[:7]
right2base_start = ee_state[7:]
draw_pose(left2base_start, s=0.1, client_id=rh5.client_id)
draw_pose(right2base_start, s=0.1, client_id=rh5.client_id)

Y = np.empty((101, 14))
Y[:, :] = ee_state[np.newaxis]
T = np.linspace(0, 1, len(Y))
Y[:, 7] += 0.1 * np.sin(T * np.pi) - 0.1
Y[:, 8] -= 0.35 * np.sin(T * 2 * np.pi)
Y[:, 9] += 0.35 * np.cos(T * 2 * np.pi)


for coupling_term in [ct, None]:
    # TODO reset DMP properly
    dmp = DualCartesianDMP(
        execution_time=execution_time, dt=dt,
        n_weights_per_dim=10, int_dt=dt, p_gain=0.0)
    dmp.imitate(T, Y)
    dmp.configure(start_y=Y[0], goal_y=Y[-1])

    #import time
    #if coupling_term is not None:
    #    time.sleep(10)
    rh5.goto_ee_state(Y[0])
    desired_positions, positions, desired_velocities, velocities = \
        rh5.step_through_cartesian(dmp, Y[0], np.zeros(12), execution_time, coupling_term=coupling_term)
    P = np.asarray(positions)
    dP = np.asarray(desired_positions)
    V = np.asarray(velocities)
    dV = np.asarray(desired_velocities)

    if plot:
        for subplot_idx, plot_dim in enumerate(range(14)):
            plt.subplot(3, 7, 1 + subplot_idx)
            plt.plot(T, Y[:, plot_dim], label="Demo %d" % plot_dim, c="k")
            plt.scatter([[0, T[-1]]], [[Y[0, plot_dim], Y[-1, plot_dim]]], c="k")
            plt.plot(np.linspace(0, execution_time, len(P)), P[:, plot_dim], label="Actual %d" % plot_dim, c="r")
            plt.scatter([[0, execution_time]], [[P[0, plot_dim], P[-1, plot_dim]]], c="r")
            plt.plot(np.linspace(0, execution_time, len(dP)), dP[:, plot_dim], label="Desired %d" % plot_dim, c="g", ls="--")
            plt.scatter([[0, execution_time]], [[dP[0, plot_dim], dP[-1, plot_dim]]], c="g")
            plt.ylim((min(P[:, plot_dim] - 0.05), max(P[:, plot_dim]) + 0.05))
        for subplot_idx, plot_dim in enumerate(range(7)):
            plt.subplot(3, 7, 15 + subplot_idx)
            plt.plot(T, Y[:, plot_dim] - Y[:, 7 + plot_dim], label="Demo %d" % plot_dim, c="k")
            plt.plot(np.linspace(0, execution_time, len(P)), P[:, plot_dim] - P[:, 7 + plot_dim], label="Actual %d" % plot_dim, c="r")
            plt.plot(np.linspace(0, execution_time, len(dP)), dP[:, plot_dim] - dP[:, 7 + plot_dim], label="Desired %d" % plot_dim, c="g", ls="--")
            D = P[:, plot_dim] - P[:, 7 + plot_dim]
            plt.ylim((min(D - 0.05), max(D) + 0.05))
        plt.legend()
        plt.show()

rh5.sim_loop()



from functools import partial
import numpy as np
from movement_primitives.dmp import CartesianDMP
import timeit


start_y = np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0])
goal_y = np.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
dt = 0.01
int_dt = 0.001

dmp = CartesianDMP(execution_time=1.0, dt=dt, n_weights_per_dim=6, int_dt=int_dt)
dmp.configure(start_y=start_y, goal_y=goal_y)
dmp.set_weights(1000 * np.random.randn(*dmp.get_weights().shape))

times = timeit.repeat(partial(dmp.open_loop, quaternion_step_function="cython"), repeat=10, number=1)
print("RK4 + Cython")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))

times = timeit.repeat(partial(dmp.open_loop, quaternion_step_function="python"), repeat=10, number=1)
print("RK4 + Python")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))

times = timeit.repeat(partial(dmp.open_loop, step_function="euler", quaternion_step_function="python"), repeat=10, number=1)
print("Euler + Python")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))



"""Kinematics
==========

Forward kinematics and a simple implementation of inverse kinematics."""
import math
import numpy as np
import numba
from pytransform3d.urdf import UrdfTransformManager
from scipy.optimize import minimize


class FastUrdfTransformManager(UrdfTransformManager):
    """Fast transformation manager that can load URDF files.

    This version has efficient numba-accelerated code to update joints.
    """
    def __init__(self):
        super(FastUrdfTransformManager, self).__init__(check=False)

    def set_joint(self, joint_name, value):
        """Set joint position.

        Note that joint values are clipped to their limits.

        Parameters
        ----------
        joint_name : string
            Name of the joint

        value : float
            Joint angle in radians in case of revolute joints or position
            in case of prismatic joint.
        """
        from_frame, to_frame, child2parent, axis, limits, joint_type = \
            self._joints[joint_name]
        value = min(max(value, limits[0]), limits[1])
        if joint_type == "revolute":
            joint2A = _fast_matrix_from_axis_angle(axis, value)
        else:
            joint2A = np.eye(4)
            joint2A[:3, 3] = value * axis
        self.transforms[(from_frame, to_frame)] = child2parent.dot(joint2A)

    def get_ee2base(self, ee_index, base_index):
        """Request a transform.

        Parameters
        ----------
        ee_index : int
            Index of the end-effector node

        base_index : int
            Index of the base node

        Returns
        -------
        ee2base : array-like, shape (4, 4)
            Homogeneous matrix that represents the transform from ee to base
        """
        return self._path_transform(self._shortest_path(ee_index, base_index))


@numba.jit(nopython=True, cache=True)
def _fast_matrix_from_axis_angle(axis, angle):  # pragma: no cover
    """Compute transformation matrix from axis-angle.

    Parameters
    ----------
    axis : array-like, shape (3,)
        Axis of rotation: (x, y, z).

    angle : float
        Rotation angle.

    Returns
    -------
    A2B : array-like, shape (4, 4)
        Transformation matrix
    """
    ux, uy, uz = axis
    c = math.cos(angle)
    s = math.sin(angle)
    ci = 1.0 - c
    ciux = ci * ux
    ciuy = ci * uy
    ciuz = ci * uz
    return np.array([
        [ciux * ux + c, ciux * uy - uz * s, ciux * uz + uy * s, 0.0],
        [ciuy * ux + uz * s, ciuy * uy + c, ciuy * uz - ux * s, 0.0],
        [ciuz * ux - uy * s, ciuz * uy + ux * s, ciuz * uz + c, 0.0],
        [0.0, 0.0, 0.0, 1.0]
    ])


class Kinematics:
    """Robot kinematics.

    Parameters
    ----------
    urdf : str
        URDF description of a robot

    mesh_path : str, optional (default: None)
        Path in which we search for meshes that are defined in the URDF.
        Meshes will be ignored if it is set to None.
    """
    def __init__(self, urdf, mesh_path=None):
        self.tm = FastUrdfTransformManager()
        self.tm.load_urdf(urdf, mesh_path=mesh_path)

    def create_chain(self, joint_names, base_frame, ee_frame, verbose=0):
        """Create kinematic chain.

        Parameters
        ----------
        joint_names : list
            Names of joints that should be used

        base_frame : str
            Name of the base link

        ee_frame : str
            Name of the end-effector link

        verbose : int, optional (default: 0)
            Verbosity level

        Returns
        -------
        chain : Chain
            Kinematic chain
        """
        return Chain(self.tm, joint_names, base_frame, ee_frame, verbose)


class Chain:
    """Kinematic chain.

    You should not call this constructor manually. Use :class:`Kinematics` to
    create a kinematic chain.

    Parameters
    ----------
    tm : FastUrdfTransformManager
        Transformation manager

    joint_names : list
        Names of joints that should be used

    base_frame : str
        Name of the base link

    ee_frame : str
        Name of the end-effector link

    verbose : int, optional (default: 0)
        Verbosity level
    """
    def __init__(self, tm, joint_names, base_frame, ee_frame, verbose=0):
        self.tm = tm
        self.joint_names = joint_names
        self.base_frame = base_frame
        self.ee_frame = ee_frame
        self.verbose = verbose

        self.joint_limits = np.array(
            [self.tm._joints[jn][4] for jn in self.joint_names])
        for i in range(len(self.joint_limits)):
            if np.isinf(self.joint_limits[i, 0]):
                self.joint_limits[i, 0] = -math.pi
            if np.isinf(self.joint_limits[i, 1]):
                self.joint_limits[i, 1] = math.pi

        self.n_joints = len(self.joint_names)
        assert len(self.joint_limits) == self.n_joints

        self.ee_index = self.tm.nodes.index(ee_frame)
        self.base_index = self.tm.nodes.index(base_frame)

    def forward(self, joint_angles):
        """Forward kinematics.

        Parameters
        ----------
        joint_angles : array, shape (n_joints,)
            Joint angles

        Returns
        -------
        ee2base : array, shape (4, 4)
            Transformation from end-effector to base frame
        """
        for i in range(self.n_joints):
            self.tm.set_joint(self.joint_names[i], joint_angles[i])
        return self.tm.get_ee2base(self.ee_index, self.base_index)

    def ee_pose_error(self, joint_angles, desired_pose, orientation_weight=1.0,
                      position_weight=1.0):
        """Compute pose error.

        Parameters
        ----------
        joint_angles : array-like, shape (n_joints,)
            Actual joint angles for which we compute forward kinematics.

        desired_pose : array-like, shape (4, 4)
            Desired pose.

        orientation_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the orientation error.

        position_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the position error.

        Returns
        -------
        pose_error : float
            Weighted error between actual pose and desired pose.
        """
        return pose_dist(desired_pose, self.forward(joint_angles),
                         orientation_weight, position_weight)

    def inverse(
            self, desired_pose, initial_joint_angles, return_error=False,
            bounds=None, solver="SLSQP", orientation_weight=1.0,
            position_weight=1.0):
        """Inverse kinematics.

        Parameters
        ----------
        desired_pose : array, shape (4, 4)
            Desired transformation from end-effector to base frame

        initial_joint_angles : array, shape (n_joints,)
            Initial guess for joint angles

        return_error : bool, optional (default: False)
            Return error in addition to joint angles

        bounds : array, shape (n_joints, 2), optional (default: joint limits)
            Bounds for joint angle optimization

        solver : str, optional (default: 'SLSQP')
            Optimizer to solve inverse kinematics problem. Possible options:
            'SLSQP', 'L-BFGS-B', and 'COBYLA'.

        orientation_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the orientation error.

        position_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the position error.

        Returns
        -------
        joint_angles : array, shape (n_joints,)
            Solution

        error : float, optional
            Pose error
        """
        if bounds is None:
            bounds = self.joint_limits
        res = minimize(
            self.ee_pose_error, initial_joint_angles,
            (desired_pose, orientation_weight, position_weight),
            method=solver, bounds=bounds)

        if self.verbose >= 2:
            print("Error: %g" % res["fun"])
        if return_error:
            return res["x"], res["fun"]
        else:
            return res["x"]

    def inverse_with_random_restarts(
            self, desired_pose, n_restarts=10, tolerance=1e-3,
            random_state=np.random, solver="SLSQP",
            orientation_weight=1.0, position_weight=1.0):
        """Compute inverse kinematics with multiple random restarts.

        Parameters
        ----------
        desired_pose : array-like, shape (4, 4)
            Desired pose.

        n_restarts : int, optional (default: 10)
            Maximum number of allowed restarts.

        tolerance : float, optional (default: 1e-3)
            Required tolerance to abort.

        random_state : np.random.RandomState, optional (default: np.random)
            Random state.

        solver : str, optional (default: 'SLSQP')
            Optimizer to solve inverse kinematics problem. Possible options:
            'SLSQP', 'L-BFGS-B', and 'COBYLA'.

        orientation_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the orientation error.

        position_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the position error.

        Returns
        -------
        joint_angles : array, shape (n_joints,)
            Solution
        """
        assert n_restarts >= 1
        Q = []
        errors = []
        for _ in range(n_restarts):
            q, error = self.inverse(
                desired_pose, self._sample_joints_uniform(random_state),
                return_error=True, solver=solver,
                orientation_weight=orientation_weight,
                position_weight=position_weight)
            Q.append(q)
            errors.append(error)
            if error <= tolerance:
                break
        if self.verbose:
            print(np.round(errors, 4))
        return Q[np.argmin(errors)]

    def local_inverse_with_random_restarts(
            self, desired_pose, joint_angles, interval, n_restarts=10,
            tolerance=1e-3, random_state=np.random, solver="SLSQP",
            orientation_weight=1.0, position_weight=1.0):
        """Compute inverse kinematics with multiple random restarts.

        Parameters
        ----------
        desired_pose : array-like, shape (4, 4)
            Desired pose.

        joint_angles : array-like, shape (n_joints,)
            Initial guess for joint angles.

        interval : float
            We will search for a solution within the range
            [joint_angles - interval, joint_angles + interval].

        n_restarts : int, optional (default: 10)
            Maximum number of allowed restarts.

        tolerance : float, optional (default: 1e-3)
            Required tolerance to abort.

        random_state : np.random.RandomState, optional (default: np.random)
            Random state.

        solver : str, optional (default: 'SLSQP')
            Optimizer to solve inverse kinematics problem. Possible options:
            'SLSQP', 'L-BFGS-B', and 'COBYLA'.

        orientation_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the orientation error.

        position_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the position error.

        Returns
        -------
        joint_angles : array, shape (n_joints,)
            Solution
        """
        assert n_restarts >= 1
        Q = []
        errors = []
        bounds = np.empty((self.n_joints, 2))
        bounds[:, 0] = joint_angles - interval
        bounds[:, 1] = joint_angles + interval
        q = joint_angles  # start with previous state
        for _ in range(n_restarts):
            q, error = self.inverse(
                desired_pose, q, return_error=True, solver=solver,
                orientation_weight=orientation_weight,
                position_weight=position_weight)
            Q.append(q)
            errors.append(error)
            if error <= tolerance:
                break
            q = self._sample_joints_uniform(random_state, bounds=bounds)
        return Q[np.argmin(errors)]

    def _sample_joints_uniform(self, random_state, bounds=None):
        if bounds is None:
            bounds = self.joint_limits
        return (random_state.rand(len(bounds)) * (bounds[:, 1] - bounds[:, 0])
                + bounds[:, 0])

    def forward_trajectory(self, Q):
        """Compute forward kinematics for a trajectory.

        Parameters
        ----------
        Q : array-like, shape (n_steps, n_joints)
            Joint angles.

        Returns
        -------
        H : array, shape (n_steps, 4, 4)
            End-effector poses.
        """
        H = np.empty((len(Q), 4, 4))
        for t in range(len(Q)):
            H[t] = self.forward(Q[t])
        return H

    def inverse_trajectory(
            self, H, initial_joint_angles=None, interval=0.1 * math.pi,
            random_restarts=True, random_state=np.random, solver="SLSQP",
            orientation_weight=1.0, position_weight=1.0):
        """Compute inverse kinematics for a trajectory.

        Parameters
        ----------
        H : array-like, shape (n_steps, 4, 4)
            Desired end-effector poses.

        initial_joint_angles : array-like, shape (n_joints,), optional (default: None)
            Initial guess for joint angles.

        interval : float
            We will search for a solution within the range
            [joint_angles - interval, joint_angles + interval] in each step.

        random_restarts : bool, optional (default: True)
            Allow random restarts if no solution is found.

        random_state : np.random.RandomState, optional (default: np.random)
            Random state.

        solver : str, optional (default: 'SLSQP')
            Optimizer to solve inverse kinematics problem. Possible options:
            'SLSQP', 'L-BFGS-B', and 'COBYLA'.

        orientation_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the orientation error.

        position_weight : float, optional (default: 1.0)
            Should be between 0.0 and 1.0 and represent the weighting for
            minimizing the position error.

        Returns
        -------
        Q : array, shape (n_steps, n_joints)
            Solution
        """
        Q = np.empty((len(H), len(self.joint_names)), dtype=float)

        if initial_joint_angles is None:
            Q[0] = self.inverse_with_random_restarts(
                H[0], random_state=random_state,
                orientation_weight=orientation_weight,
                position_weight=position_weight)
        else:
            Q[0] = self.inverse(
                H[0], initial_joint_angles, solver=solver,
                orientation_weight=orientation_weight,
                position_weight=position_weight)

        for t in range(1, len(H)):
            if self.verbose >= 2:
                print("Step: %d" % (t + 1))
            if random_restarts:
                Q[t] = self.local_inverse_with_random_restarts(
                    H[t], Q[t - 1], interval, random_state=random_state,
                    orientation_weight=orientation_weight,
                    position_weight=position_weight)
            else:
                bounds = np.empty((self.n_joints, 2), dtype=float)
                bounds[:, 0] = Q[t - 1] - interval
                bounds[:, 1] = Q[t - 1] + interval
                Q[t] = self.inverse(
                    H[t], Q[t - 1], False, bounds, solver=solver,
                    orientation_weight=orientation_weight,
                    position_weight=position_weight)
        return Q


@numba.jit(nopython=True, cache=True)
def pose_dist(
        ee2base_desired, ee2base_actual, orientation_weight, position_weight):  # pragma: no cover
    ee_actual2ee_desired = np.linalg.inv(ee2base_actual).dot(ee2base_desired)
    trace = (ee_actual2ee_desired[0, 0] + ee_actual2ee_desired[1, 1]
             + ee_actual2ee_desired[2, 2])
    angle = math.acos(min((trace - 1.0) / 2.0, 1.0))
    orientation_error = min(angle, 2.0 * math.pi - angle)
    position_error = math.sqrt(
        ee_actual2ee_desired[0, 3] ** 2 + ee_actual2ee_desired[1, 3] ** 2
        + ee_actual2ee_desired[2, 3] ** 2)
    return (orientation_weight * orientation_error
            + position_weight * position_error)



import numpy as np


def canonical_system_alpha(goal_z, goal_t, start_t):
    r"""Compute parameter alpha of canonical system.

    The parameter alpha is computed such that a specific phase value goal_z
    is reached at goal_t. The canonical system is defined according to [1]_,
    even though we compute a different value for alpha.

    Parameters
    ----------
    goal_z : float
        Value of phase variable at the end of the execution (> 0).

    goal_t : float
        Time at which the execution should be done. Make sure that
        goal_t > start_t.

    start_t : float
        Time at which the execution should start.

    Returns
    -------
    alpha : float
        Value of the alpha parameter of the canonical system.

    Raises
    ------
    ValueError
        If input values are invalid.

    References
    ----------
    .. [1] Ijspeert, A. J., Nakanishi, J., Hoffmann, H., Pastor, P., Schaal, S.
       (2013). Dynamical Movement Primitives: Learning Attractor Models for
       Motor Behaviors. Neural Computation 25 (2), 328-373. DOI:
       10.1162/NECO_a_00393,
       https://homes.cs.washington.edu/~todorov/courses/amath579/reading/DynamicPrimitives.pdf
    """
    if goal_z <= 0.0:
        raise ValueError("Final phase must be > 0!")
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    return float(-np.log(goal_z))


def phase(t, alpha, goal_t, start_t):
    r"""Map time to phase.

    According to [1]_, the differential Equation

    .. math::

        \tau \dot{z} = -\alpha_z z

    describes the evolution of the phase variable z. Starting from the initial
    position :math:`z_0 = 1`, the phase value converges monotonically to 0.
    Instead of using an iterative procedure to calculate the current value of
    z, it is computed directly through

    .. math::

        z(t) = \exp - \frac{\alpha_z}{\tau} t

    Parameters
    ----------
    t : float
        Current time.

    alpha : float
        Value of the alpha parameter of the canonical system.

    goal_t : float
        Time at which the execution should be done.

    start_t : float
        Time at which the execution should start.

    Returns
    -------
    z : float
        Value of phase variable.

    References
    ----------
    .. [1] Ijspeert, A. J., Nakanishi, J., Hoffmann, H., Pastor, P., Schaal, S.
       (2013). Dynamical Movement Primitives: Learning Attractor Models for
       Motor Behaviors. Neural Computation 25 (2), 328-373. DOI:
       10.1162/NECO_a_00393,
       https://homes.cs.washington.edu/~todorov/courses/amath579/reading/DynamicPrimitives.pdf
    """
    execution_time = goal_t - start_t
    return np.exp(-alpha * (t - start_t) / execution_time)
